{ File Association for all Windows versions.
  This includes Windows Vista/7/8 Default Programs feature.
  Author: Lainz. }
unit FileAssociation;

{$mode objfpc}{$H+}

interface

uses
  Classes, Registry{$ifdef windows}, ShlObj{$endif};

const
  HKCR = HKEY_CLASSES_ROOT;
  HKCU = HKEY_CURRENT_USER;
  HKLM = HKEY_LOCAL_MACHINE;
  HKU = HKEY_USERS;
  HKCC = HKEY_CURRENT_CONFIG;

type
  TAppReg = record
    AppName,        // 'Lazarus IDE'
    AppDescription, // 'Open Source IDE for Free Pascal'
    AppNameNoSpaces,// 'LazarusIDE'
    ExtData,        // '.lpr'
    ExtIcon,        // 'C:\FPC\Lazarus\images\lazaruspackage.ico'
    ExtName,        // 'Lazarus package source file'
    ExtNameNoSpaces,// 'LazarusPackageSourceFile'
    CmdData,        // '"C:\FPC\Lazarus\lazarus.exe" "%1"'
    CmdIcon,        // 'C:\FPC\Lazarus\lazarus.exe,0'
    CmdName,        // 'Open Project'
    CmdNameNoSpaces // 'Open'
    : string;
  end;

  { TFileAssociation }

  TFileAssociation = class(TComponent)
  private
    FAppName, FAppDescription, FAppNameNoSpaces, FExtData, FExtIcon,
    FExtName, FExtNameNoSpaces, FCmdData, FCmdIcon, FCmdName, FCmdNameNoSpaces: string;
    FRegisterAssociation, FRegisterForAllUsers, FAddAppToDefaultPrograms,
    FAddExtToDefaultPrograms: boolean;
    FUnReg: boolean;
    procedure SetFAddAppToDefaultProgams(AValue: boolean);
    procedure SetFAddExtToDefaultProgams(AValue: boolean);
    procedure SetFAppDescription(AValue: string);
    procedure SetFAppName(AValue: string);
    procedure SetFAppNameNoSpaces(AValue: string);
    procedure SetFCmdData(AValue: string);
    procedure SetFCmdIcon(AValue: string);
    procedure SetFCmdName(AValue: string);
    procedure SetFCmdNameNoSpaces(AValue: string);
    procedure SetFExtData(AValue: string);
    procedure SetFExtIcon(AValue: string);
    procedure SetFExtName(AValue: string);
    procedure SetFExtNameNoSpaces(AValue: string);
    procedure SetFRegisterAssociation(AValue: boolean);
    procedure SetFRegisterForAllUsers(AValue: boolean);
    procedure SetFUnReg(AValue: boolean);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Execute: boolean;
    procedure ClearIconCache;
  published
    property AppName: string read FAppName write SetFAppName;
    property AppDescription: string read FAppDescription write SetFAppDescription;
    property AppNameNoSpaces: string read FAppNameNoSpaces write SetFAppNameNoSpaces;
    property ExtData: string read FExtData write SetFExtData;
    property ExtIcon: string read FExtIcon write SetFExtIcon;
    property ExtName: string read FExtName write SetFExtName;
    property ExtNameNoSpaces: string read FExtNameNoSpaces write SetFExtNameNoSpaces;
    property CmdData: string read FCmdData write SetFCmdData;
    property CmdIcon: string read FCmdIcon write SetFCmdIcon;
    property CmdName: string read FCmdName write SetFCmdName;
    property CmdNameNoSpaces: string read FCmdNameNoSpaces write SetFCmdNameNoSpaces;
    property RegisterAssociation: boolean read FRegisterAssociation
      write SetFRegisterAssociation;
    property RegisterForAllUsers: boolean read FRegisterForAllUsers
      write SetFRegisterForAllUsers;
    property AddAppToDefaultPrograms: boolean
      read FAddAppToDefaultPrograms write SetFAddAppToDefaultProgams;
    property AddExtToDefaultPrograms: boolean
      read FAddExtToDefaultPrograms write SetFAddExtToDefaultProgams;
    property UnReg: boolean read FUnReg write SetFUnReg default False;
  end;

function WriteStringValue(Root: HKEY; SubKey: string; ValueName: string;
  ValueData: string): boolean;
function DeleteValue(Root: HKEY; SubKey: string; ValueName: string): boolean;

{ Write a string to Root\SubKey\ ValueName=ValueData }
function WriteString(Root: HKEY; SubKey: string; ValueName: string;
  ValueData: string; UnReg: boolean = False): boolean;

{ Registry 'Class' containing the icon }
function WriteFileAssociationClass(Root: HKEY;
  AppNameNoSpaces, ExtIcon, ExtName, ExtNameNoSpaces: string;
  UnReg: boolean = False): boolean;
function WriteFileAssociationClass(Root: HKEY; app: TAppReg;
  UnReg: boolean = False): boolean;

{ Add a command like 'Open', 'Edit', 'Print' or other }
function WriteFileAssociationClassCommand(Root: HKEY;
  AppNameNoSpaces, ExtNameNoSpaces, CmdName, CmdNameNoSpaces, CmdIcon, CmdData: string;
  UnReg: boolean = False): boolean;
function WriteFileAssociationClassCommand(Root: HKEY; app: TAppReg;
  UnReg: boolean = False): boolean;

{ Associate the 'Class' with the file extension }
function WriteFileAssociation(Root: HKEY;
  AppNameNoSpaces, ExtNameNoSpaces, ExtData: string; UnReg: boolean = False): boolean;
function WriteFileAssociation(Root: HKEY; app: TAppReg; UnReg: boolean = False): boolean;

{ Add our application to Default Programs (Vista+) }
function WriteDefaultPrograms(AppNameNoSpaces, AppName, AppDescription: string;
  UnReg: boolean = False): boolean;
function WriteDefaultPrograms(app: TAppReg; UnReg: boolean = False): boolean;

{ Add extension to our application Default Programs page }
function WriteDefaultProgramsAddExt(AppNameNoSpaces, ExtData, ExtNameNoSpaces: string;
  UnReg: boolean = False): boolean;
function WriteDefaultProgramsAddExt(app: TAppReg; UnReg: boolean = False): boolean;

{ Create a File Association with all the previous functions together }
{ RegisterAssociation = True = this do the real association; = False = only the 'Class' will be created }
{ RegisterForAllUsers = True = this uses HKLM; = False uses HKCU and does not AddToDefaultPrograms}
{ AddAppToDefaultPrograms = True = add app to default programs in control panel, it only need to be called one time}
{ AddExtToDefaultPrograms = True = add ext to default programs in control panel, it needs AddAppToDefaultPrograms at least one time}
function CreateFileAssociation(app: TAppReg;
  RegisterAssociation, RegisterForAllUsers, AddAppToDefaultPrograms,
  AddExtToDefaultPrograms: boolean; UnReg: boolean = False): boolean;

{ Update icons }
procedure ClearIconCache;

procedure Register;

implementation

procedure Register;
begin
  //{$I fileassociation_icon.lrs}
  RegisterComponents('System', [TFileAssociation]);
end;

function WriteString(Root: HKEY; SubKey: string; ValueName: string;
  ValueData: string; UnReg: boolean = False): boolean;
begin
  if not UnReg then
    Result := WriteStringValue(Root, SubKey, ValueName, ValueData)
  else
    Result := DeleteValue(Root, SubKey, ValueName);
end;

function WriteStringValue(Root: HKEY; SubKey: string; ValueName: string;
  ValueData: string): boolean;
var
  reg: TRegistry;
begin
  reg := TRegistry.Create;
  reg.RootKey := Root;

  if reg.OpenKey(SubKey, True) then
  begin
    reg.WriteString(ValueName, ValueData);
    reg.CloseKey;
    reg.Free;
    Result := True;
  end
  else
    Result := False;
end;

function DeleteValue(Root: HKEY; SubKey: string; ValueName: string): boolean;
var
  reg: TRegistry;
begin
  reg := TRegistry.Create;
  reg.RootKey := Root;

  if reg.OpenKey(SubKey, True) then
  begin
    reg.DeleteKey(ValueName);
    reg.CloseKey;
    reg.Free;
    Result := True;
  end
  else
    Result := False;
end;

function WriteFileAssociationClass(Root: HKEY;
  AppNameNoSpaces, ExtIcon, ExtName, ExtNameNoSpaces: string;
  UnReg: boolean = False): boolean;
var
  r1, r2: boolean;
begin
  r1 := WriteString(Root, concat('Software\Classes\', AppNameNoSpaces,
    '.AssocFile.', ExtNameNoSpaces), '', ExtName, UnReg);
  r2 := WriteString(Root, concat('Software\Classes\', AppNameNoSpaces,
    '.AssocFile.', ExtNameNoSpaces, '\DefaultIcon'), '', ExtIcon, UnReg);

  if (r1 = False) or (r2 = False) then
    Result := False
  else
    Result := True;
end;

function WriteFileAssociationClass(Root: HKEY; app: TAppReg;
  UnReg: boolean = False): boolean;
begin
  Result := WriteFileAssociationClass(Root, app.AppNameNoSpaces,
    app.ExtIcon, app.ExtName, app.ExtNameNoSpaces, UnReg);
end;

function WriteFileAssociationClassCommand(Root: HKEY;
  AppNameNoSpaces, ExtNameNoSpaces, CmdName, CmdNameNoSpaces, CmdIcon, CmdData: string;
  UnReg: boolean = False): boolean;
var
  r1, r2, r3: boolean;
begin
  r1 := WriteString(Root, concat('Software\Classes\', AppNameNoSpaces,
    '.AssocFile.', ExtNameNoSpaces, '\Shell\', CmdNameNoSpaces), '', CmdName, UnReg);
  r2 := WriteString(Root, concat('Software\Classes\', AppNameNoSpaces,
    '.AssocFile.', ExtNameNoSpaces, '\Shell\', CmdNameNoSpaces), 'Icon', CmdIcon, UnReg);
  r3 := WriteString(Root, concat('Software\Classes\', AppNameNoSpaces,
    '.AssocFile.', ExtNameNoSpaces, '\Shell\', CmdNameNoSpaces, '\Command'),
    '', CmdData, UnReg);

  if (r1 = False) or (r2 = False) or (r3 = False) then
    Result := False
  else
    Result := True;
end;

function WriteFileAssociationClassCommand(Root: HKEY; app: TAppReg;
  UnReg: boolean = False): boolean;
begin
  Result := WriteFileAssociationClassCommand(Root, app.AppNameNoSpaces,
    app.ExtNameNoSpaces, app.CmdName, app.CmdNameNoSpaces, app.CmdIcon,
    app.CmdData, UnReg);
end;

function WriteFileAssociation(Root: HKEY;
  AppNameNoSpaces, ExtNameNoSpaces, ExtData: string; UnReg: boolean = False): boolean;
begin
  Result := WriteString(Root, concat('Software\Classes\', ExtData),
    '', concat(AppNameNoSpaces, '.AssocFile.', ExtNameNoSpaces));
end;

function WriteFileAssociation(Root: HKEY; app: TAppReg; UnReg: boolean = False): boolean;
begin
  Result := WriteFileAssociation(Root, app.AppNameNoSpaces, app.ExtNameNoSpaces,
    app.ExtData, UnReg);
end;

function WriteDefaultPrograms(AppNameNoSpaces, AppName, AppDescription: string;
  UnReg: boolean = False): boolean;
var
  r1, r2, r3, r4: boolean;
begin
  r1 := WriteString(HKLM, concat('Software\', AppNameNoSpaces, '\Capabilities'),
    '', '', UnReg);
  r2 := WriteString(HKLM, concat('Software\', AppNameNoSpaces, '\Capabilities'),
    'ApplicationName', AppName, UnReg);
  r3 := WriteString(HKLM, concat('Software\', AppNameNoSpaces, '\Capabilities'),
    'ApplicationDescription', AppDescription, UnReg);
  r4 := WriteString(HKLM, 'Software\RegisteredApplications',
    AppNameNoSpaces, concat('Software\', AppNameNoSpaces, '\Capabilities'), UnReg);
  if (r1 = False) or (r2 = False) or (r3 = False) or (r4 = False) then
    Result := False
  else
    Result := True;
end;

function WriteDefaultPrograms(app: TAppReg; UnReg: boolean = False): boolean;
begin
  Result := WriteDefaultPrograms(app.AppNameNoSpaces, app.AppName,
    app.AppDescription, UnReg);
end;

function WriteDefaultProgramsAddExt(AppNameNoSpaces, ExtData, ExtNameNoSpaces: string;
  UnReg: boolean = False): boolean;
begin
  Result := WriteString(HKLM, concat('Software\', AppNameNoSpaces,
    '\Capabilities\FileAssociations'), ExtData,
    concat(AppNameNoSpaces, '.AssocFile.', ExtNameNoSpaces), UnReg);
end;

function WriteDefaultProgramsAddExt(app: TAppReg; UnReg: boolean = False): boolean;
begin
  Result := WriteDefaultProgramsAddExt(app.AppNameNoSpaces, app.ExtData,
    app.ExtNameNoSpaces, UnReg);
end;

function CreateFileAssociation(app: TAppReg;
  RegisterAssociation, RegisterForAllUsers, AddAppToDefaultPrograms,
  AddExtToDefaultPrograms: boolean; UnReg: boolean = False): boolean;
var
  r1, r2, r3, r4, r5: boolean;
  Root: HKEY;
begin
  if RegisterForAllUsers then
    Root := HKLM
  else
    Root := HKCU;

  r1 := WriteFileAssociationClass(Root, app, UnReg);
  r2 := WriteFileAssociationClassCommand(Root, app, UnReg);

  if RegisterAssociation then
    r3 := WriteFileAssociation(Root, app, UnReg);

  if Root <> HKCU then
  begin
    if AddAppToDefaultPrograms then
      r4 := WriteDefaultPrograms(app, UnReg);
    if AddExtToDefaultPrograms then
      r5 := WriteDefaultProgramsAddExt(app, UnReg);
  end;

  if (r1 = False) or (r2 = False) or (r3 = False) or (r4 = False) or (r5 = False) then
    Result := False
  else
    Result := True;
end;

procedure ClearIconCache;
begin
  {$ifdef windows}SHChangeNotify(SHCNE_ASSOCCHANGED, SHCNF_IDLIST, nil, nil);{$endif}
end;

{ TFileAssociation }

procedure TFileAssociation.SetFAppDescription(AValue: string);
begin
  if FAppDescription = AValue then
    Exit;
  FAppDescription := AValue;
end;

procedure TFileAssociation.SetFAddAppToDefaultProgams(AValue: boolean);
begin
  if FAddAppToDefaultPrograms = AValue then
    Exit;
  FAddAppToDefaultPrograms := AValue;
end;

procedure TFileAssociation.SetFAddExtToDefaultProgams(AValue: boolean);
begin
  if FAddExtToDefaultPrograms = AValue then
    Exit;
  FAddExtToDefaultPrograms := AValue;
end;

procedure TFileAssociation.SetFAppName(AValue: string);
begin
  if FAppName = AValue then
    Exit;
  FAppName := AValue;
end;

procedure TFileAssociation.SetFAppNameNoSpaces(AValue: string);
begin
  if FAppNameNoSpaces = AValue then
    Exit;
  FAppNameNoSpaces := AValue;
end;

procedure TFileAssociation.SetFCmdData(AValue: string);
begin
  if FCmdData = AValue then
    Exit;
  FCmdData := AValue;
end;

procedure TFileAssociation.SetFCmdIcon(AValue: string);
begin
  if FCmdIcon = AValue then
    Exit;
  FCmdIcon := AValue;
end;

procedure TFileAssociation.SetFCmdName(AValue: string);
begin
  if FCmdName = AValue then
    Exit;
  FCmdName := AValue;
end;

procedure TFileAssociation.SetFCmdNameNoSpaces(AValue: string);
begin
  if FCmdNameNoSpaces = AValue then
    Exit;
  FCmdNameNoSpaces := AValue;
end;

procedure TFileAssociation.SetFExtData(AValue: string);
begin
  if FExtData = AValue then
    Exit;
  FExtData := AValue;
end;

procedure TFileAssociation.SetFExtIcon(AValue: string);
begin
  if FExtIcon = AValue then
    Exit;
  FExtIcon := AValue;
end;

procedure TFileAssociation.SetFExtName(AValue: string);
begin
  if FExtName = AValue then
    Exit;
  FExtName := AValue;
end;

procedure TFileAssociation.SetFExtNameNoSpaces(AValue: string);
begin
  if FExtNameNoSpaces = AValue then
    Exit;
  FExtNameNoSpaces := AValue;
end;

procedure TFileAssociation.SetFRegisterAssociation(AValue: boolean);
begin
  if FRegisterAssociation = AValue then
    Exit;
  FRegisterAssociation := AValue;
end;

procedure TFileAssociation.SetFRegisterForAllUsers(AValue: boolean);
begin
  if FRegisterForAllUsers = AValue then
    Exit;
  FRegisterForAllUsers := AValue;
end;

procedure TFileAssociation.SetFUnReg(AValue: boolean);
begin
  if FUnReg = AValue then
    Exit;
  FUnReg := AValue;
end;

constructor TFileAssociation.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FRegisterAssociation := True;
  FRegisterForAllUsers := True;
  FAddAppToDefaultPrograms := True;
  FAddExtToDefaultPrograms := True;
  FUnReg := False;
end;

destructor TFileAssociation.Destroy;
begin
  inherited Destroy;
end;

function TFileAssociation.Execute: boolean;
var
  r1, r2, r3, r4, r5: boolean;
  Root: HKEY;
begin
  if RegisterForAllUsers then
    Root := HKLM
  else
    Root := HKCU;

  r1 := FileAssociation.WriteFileAssociationClass(Root, AppNameNoSpaces,
    ExtIcon, ExtName, ExtNameNoSpaces, UnReg);
  r2 := FileAssociation.WriteFileAssociationClassCommand(Root,
    AppNameNoSpaces, ExtNameNoSpaces, CmdName, CmdNameNoSpaces, CmdIcon, CmdData, UnReg);

  if RegisterAssociation then
    r3 := FileAssociation.WriteFileAssociation(Root, AppNameNoSpaces,
      ExtNameNoSpaces, ExtData, UnReg);

  if Root <> HKCU then
  begin
    if AddAppToDefaultPrograms then
      r4 := FileAssociation.WriteDefaultPrograms(AppNameNoSpaces,
        AppName, AppDescription, UnReg);
    if AddExtToDefaultPrograms then
      r5 := FileAssociation.WriteDefaultProgramsAddExt(AppNameNoSpaces,
        ExtData, ExtNameNoSpaces, UnReg);
  end;

  if (r1 = False) or (r2 = False) or (r3 = False) or (r4 = False) or (r5 = False) then
    Result := False
  else
    Result := True;
end;

procedure TFileAssociation.ClearIconCache;
begin
  FileAssociation.ClearIconCache;
end;

end.
